from functools import total_ordering
from typing import List, Union


@total_ordering
class Part:
    """Creates an instance of Part.

    Part objects represent a collection of positive integers.

    Attributes:
        unit (dict): The unit of measurement of time (see units.py).
        options (dict): The options to use: output_weekday_names and output_month_names
    """
    def __init__(self, unit, options):
        self.options = options if bool(options) else dict()
        self.unit = unit
        self.values = None

    def __str__(self) -> str:
        """Print directly the Part Object"""
        return self.to_string()

    def __repr__(self):
        return f'{self.__class__.__name__} - (values:{self.values!r}, unit:{self.unit.get("name")!r})'

    def __len__(self):
        return len(self.to_list())

    def __lt__(self, other) -> bool:
        """ This Part object is lower than the other Part.
        The comparison is made by 'total_ordering', comparing self.values list.
        """
        return len(self) < len(other)

    def __eq__(self, other) -> bool:
        """ This Part object is equal to the other Part.
        The comparison is made by 'total_ordering', comparing self.values list.
        """
        return len(self) == len(other)

    def from_list(self, part_list: List[Union[str, int]]) -> None:
        """Validates a range of positive integers.

        :param part_list: An array of positive integers.
        :raises ValueError: An error occurred in case of invalid value or out of range value
        """
        values = []
        for part_value in part_list:
            try:
                parsed_value = int(part_value)
            except ValueError:
                raise ValueError(f'Invalid list value {part_value}')
            if parsed_value is None:
                raise ValueError(f'Invalid list value {part_value}')
            values.append(parsed_value)

        if not values:
            raise ValueError('Empty interval value')

        sunday_fixed_values = self._fix_sunday(values)
        unique_values = list(dict.fromkeys(sunday_fixed_values))  # Remove eventual duplicates
        unique_values.sort()
        part_value = self.out_of_range(unique_values)
        if part_value is not None:
            raise ValueError(f'Value {part_value!r} out of range for {self.unit.get("name")!r}')

        self.values = unique_values

    def from_string(self, cron_part: str) -> None:
        """Parses a string as a range of positive integers.

        :param cron_part: The string that represent a Part. It will be converted as a range.
        :raises ValueError: Invalid value.
        :raises ValueError: An error occurred in case of invalid value or out of range value.
        """
        string_parts = cron_part.split('/')  # Split in the case of step parameter
        if len(string_parts) > 2:
            raise ValueError(f'Invalid value {cron_part}')

        range_string = self._replace_alternatives(string_parts[0])
        if not range_string:
            raise ValueError(f'Invalid value {range_string}')
        elif range_string == '*':
            unit_range = self.possible_values()
        else:
            ranges_lists = [self._parse_range(range_string_part) for range_string_part in range_string.split(',')]
            flattened_ranges_list = [item for sublist in ranges_lists for item in sublist]
            flattened_ranges_list = self._fix_sunday(flattened_ranges_list)
            unit_range = list(dict.fromkeys(flattened_ranges_list))  # Remove eventual duplicates
            unit_range.sort()
            value = self.out_of_range(unit_range)
            if value is not None:
                raise ValueError(f'Value {value!r} out of range for {self.unit.get("name")!r}')

        step = self._get_step(string_parts)

        interval_values = self._apply_interval(unit_range, step)  # filter by step
        if not len(interval_values):
            raise ValueError(f'Empty intervals value {cron_part}')

        self.values = interval_values

    def _fix_sunday(self, values: List[int]) -> List[int]:
        """Replaces all 7 with 0 as Sunday can be represented by both.

         :param values: The values to process.
         :returns: values -> The resulting array.
         """
        if self.unit.get('name') == 'weekday':
            values = [0 if value == 7 else value for value in values]
        return values

    @staticmethod
    def _parse_range(unit_range: str):
        """Parses a range string. Example "15-19"

        :param unit_range: The range string.
        :return: The resulting array.
        :raise ValueError: Impossible to convert the Part unit as int.
        :raise ValueError: Invalid min or max value for the Part unit.
        :raise ValueError: Not valid Range, max range is less than min range
        """
        sub_parts = unit_range.split('-')
        if len(sub_parts) == 1:
            try:
                value = int(sub_parts[0])
            except ValueError as exc:
                raise ValueError(f'Invalid value {unit_range!r} --> {exc}')
            return [value]
        elif len(sub_parts) == 2:
            try:
                min_value = int(sub_parts[0])
                max_value = int(sub_parts[1])
            except ValueError as exc:
                raise ValueError(f'Invalid min or max value from: {unit_range!r} --> {exc}')
            if max_value < min_value:
                raise ValueError(f'Max range is less than min range in {unit_range}')
            return [int_value for int_value in range(min_value, max_value + 1)]
        else:
            raise ValueError(f'Invalid value {unit_range}')

    def _get_step(self, string_parts: List[str]) -> Union[None, int]:
        """Get the step part of the part string.

        :param string_parts: the part string.
        :return step: parsed step.
        :raise IndexError: The second index of the list does not exists. The step is not present.
        """
        try:
            step = string_parts[1]
        except IndexError:
            step = None

        if step or step == '':
            step = self._parse_step(step)

        return step

    def _parse_step(self, step: str) -> int:
        """Parses the step from a part string.

        :param step: The step string.
        :return: The step value.
        :raise ValueError: Invalid interval step value.
        """
        try:
            parsed_step = int(step)
        except (ValueError, TypeError):
            raise ValueError(f'Invalid interval step value {step!r} for {self.unit.get("name")!r}')
        if not parsed_step or parsed_step < 1:
            raise ValueError(f'Invalid interval step value {step!r} for {self.unit.get("name")!r}')
        return parsed_step

    @staticmethod
    def _apply_interval(values: List[int], step: int) -> List[int]:
        """Applies an interval step to a collection of values.

        :param values: A collection of numbers.
        :param step: The step value.
        :returns: The resulting collection.
        """
        if step:
            min_value = values[0]
            values = [value for value in values if value % step == min_value % step or value == min_value]
        return values

    def _replace_alternatives(self, string: str) -> str:
        """Replaces the alternative representations of numbers in a string.
        Example -> month '12' will be 'DEC'

        :param string: The string to process.
        :return: The processed string.
        """
        if 'alt' in self.unit:
            string = string.upper()
            for idx, alt in enumerate(self.unit.get('alt')):
                string = string.replace(alt, str(self.unit.get("min") + idx))
        return string

    def out_of_range(self, values: List[int]) -> Union[int, None]:
        """Finds an element from values that is outside of the range of self.unit

        :param values: The values to test.
        :return: An integer is a value out of range was found, otherwise None.
        """
        first = values[0]
        last = values[-1]
        if first < self.unit.get('min'):
            return first
        elif last > self.unit.get('max'):
            return last
        else:
            return None

    def possible_values(self) -> List[int]:
        """Creates a list of Part Unit possible values from 'min' to 'max'

        :return: A List of Part Unit values
        """
        return list(range(self.unit.get('min'), self.unit.get('max') + 1))

    def min(self) -> int:
        """Returns the smallest value in the range.
        Example -> month Part: 1

        :return: The smallest Part value.
        """
        return self.values[0]

    def max(self) -> int:
        """Returns the largest value in the range.
        Example -> month Part: 12

        :return: The largest Part value.
        """
        return self.values[-1]

    def is_full(self) -> bool:
        """Returns true if range has all the values of the unit.
        """
        return len(self.values) == self.unit.get('max') - self.unit.get('min') + 1

    def get_step(self) -> Union[int, None]:
        """Returns the difference between first and second elements in the range.

        :return: step between numbers in the interval whether array interval > 2.
        """
        if self.values and len(self.values) > 2:
            step = self.values[1] - self.values[0]
            if step > 1:
                return step
        else:
            return None

    def is_interval(self, step: int) -> bool:
        """Returns true if the range can be represented as an interval.

        :param step: The difference between numbers in the interval.
        """
        for idx, value in enumerate(self.values):
            if self.values[0] == value:
                continue
            prev_value = self.values[idx - 1]
            current_value = value
            if current_value - prev_value != step:
                return False
        return True

    def is_full_interval(self, step: int) -> bool:
        """Returns true if the range contains all the interval values.

        :param step: The difference between numbers in the interval.
        """
        have_all_values = len(self.values) == round((self.max() - self.min()) / step) + 1
        if self.min() == self.unit.get('min') and self.max() + step > self.unit.get('max') and have_all_values:
            return True
        return False

    def has(self, value: int) -> bool:
        """Checks if the range contains the specified value.

        :param value: The value to look for.
        :return: Whether the value is present in the range or not.
        """
        try:
            self.values.index(value)
            return True
        except ValueError:
            return False

    def to_list(self) -> List[int]:
        """Returns the range as an array of positive integers.

        :return: The range as an array.
        """
        return self.values

    def to_ranges(self) -> List[List[Union[str, int]]]:
        """Returns the range as an array of ranges defined as arrays of positive integers.

        :return: multi_dim_values (list of list): The range as a multi-dimensional array.
        """
        multi_dim_values = list()
        start_number = None
        for idx, value in enumerate(self.values):
            try:
                next_value = self.values[idx + 1]
            except IndexError:
                next_value = -1  # No next item in the self.values list

            if value != next_value - 1:  # next_value is not the subsequent number
                if start_number is not None:
                    multi_dim_values.append([start_number, value])
                    start_number = None
                else:  # The last number of the list "self.values" is not in a range
                    multi_dim_values.append(value)
            elif start_number is None:
                start_number = value

        return multi_dim_values

    def to_string(self) -> str:
        """Returns the range as a string.

        :return: The range as a string.
        """
        cron_range_strings = []
        if self.is_full():
            if 'output_hashes' in self.options:
                cron_part_str = 'H'
            else:
                cron_part_str = '*'
        else:
            step = self.get_step()
            if step and self.is_interval(step):
                if self.is_full_interval(step):
                    if 'output_hashes' in self.options:
                        cron_part_str = f'H/{step}'
                    else:
                        cron_part_str = f'*/{step}'
                else:
                    if 'output_hashes' in self.options:
                        cron_part_str = f'H({self.format_value(self.min())}-{self.format_value(self.max())})/{step}'
                    else:
                        cron_part_str = f'{self.format_value(self.min())}-{self.format_value(self.max())}/{step}'
            else:
                cron_ranges = self.to_ranges()
                for cron_range in cron_ranges:
                    if isinstance(cron_range, list):
                        cron_range_strings.append(
                            f'{self.format_value(cron_range[0])}-{self.format_value(cron_range[1])}')
                    else:
                        cron_range_strings.append(f'{self.format_value(cron_range)}')

                if isinstance(cron_range_strings, list) and cron_range_strings:
                    cron_part_str = ','.join(cron_range_strings)
                else:
                    cron_part_str = cron_range_strings[0]

        return cron_part_str

    def format_value(self, value: int) -> Union[int, str]:
        """Formats weekday and month names as string when the relevant options are set.

        :param value: The value to process.
        :return: The formatted string or number.
        """
        if 'output_weekday_names' in self.options and self.options.get('output_weekday_names') and \
                self.unit.get('name') == 'weekday' or 'output_month_names' in self.options and \
                self.options.get('output_month_names') and self.unit.get('name') == 'month':
            return_value = self.unit.get("alt")[value - self.unit.get('min')]
        else:
            return_value = value
        return return_value



from datetime import datetime, timedelta, timezone
from dateutil import tz
import calendar
import copy
from typing import Optional

from typing import TYPE_CHECKING, List, Literal

if TYPE_CHECKING:
    from cron import Cron
    from sub_modules.part import Part

weekdays = {'Sun': 0, 'Mon': 1, 'Tue': 2, 'Wed': 3, 'Thu': 4, 'Fri': 5, 'Sat': 6}  # en_US weekdays


class Seeker:
    """Create an instance of Seeker. Seeker objects search for execution times of a cron schedule.
    Args:
        cron (object): Cron object
        start_date (datetime): The start date for the schedule iterator, with or without timezone.
        timezone_str (str): The timezone to make an timezone aware datetime as response.
    """
    def __init__(self, cron: 'Cron', start_date: Optional[datetime] = None, timezone_str: Optional[str] = None) -> None:
        if not cron.parts:
            raise LookupError('No schedule found')

        if start_date is not None and timezone_str is not None:
            raise ValueError('should have location_num or location_path, but not both')
        if start_date:
            # Construct the Seeker object from a past or a future date
            try:
                isinstance(start_date, datetime)
                self.tz_info = start_date.tzinfo
                self.date = start_date
            except Exception as exc:
                raise ValueError(f'Input schedule start time is not a valid datetime object. Error -> {exc}')
        elif timezone_str:
            if tz.gettz(timezone_str):
                self.tz_info = tz.gettz(timezone_str)
                self.date = datetime.now(self.tz_info)
            else:
                raise ValueError(f'Provided not a valid Timezone --> {timezone_str}')
        else:
            self.date = datetime.now(tz.tzutc())

        if self.date.second > 0:
            # Add a minute to the date to prevent returning dates in the past
            self.date = self.date + timedelta(minutes=+1)

        self.start_time = self.date
        self.cron = cron
        self.pristine = True

    """Resets the iterator."""
    def reset(self) -> None:
        self.pristine = True
        self.date = self.start_time

    """Returns the time the schedule would run next.
    
     Returns:
        (datetime): The time the schedule would run next.
    """
    def next(self) -> datetime:
        if self.pristine:
            self.pristine = False
        else:
            one_minute = timedelta(minutes=+1)
            # Ensure next is never now
            self.date = self.date + one_minute

        return self.find_date(getattr(self.cron, 'parts'))

    """Returns the time the schedule would have last run at.
    
    Returns:
        (datetime): The time the schedule would have last run at.
    """
    def prev(self) -> datetime:
        self.pristine = False
        # Ensure prev and next cannot be same time
        self.date = self.date + timedelta(minutes=-1)
        return self.find_date(getattr(self.cron, 'parts'), True)

    """Returns the time the schedule would run next. # TODO refactor description.
    
    Args:
        cron_parts (List): List of all cron 'Part'.
        reverse(boolean): Whether to find the previous value instead of next.
    Returns:
        (datetime): A new datetime object. The date the schedule would have executed at.
    """
    def find_date(self, cron_parts: List['Part'], reverse: bool = False) -> datetime:
        operation: Literal['add', 'subtract'] = 'add'
        if reverse:
            operation = 'subtract'
        retry = 24
        while retry:
            retry -= 1
            self._shift_month(cron_parts[3], operation)
            month_changed = self._shift_day(cron_parts[2], cron_parts[4], operation)
            if not month_changed:
                day_changed = self._shift_hour(cron_parts[1], operation)
                if not day_changed:
                    hour_changed = self._shift_minute(cron_parts[0], operation)
                    if not hour_changed:
                        break
        else:
            raise Exception('Unable to find execution time for schedule')

        return copy.deepcopy(self.date.replace(second=0, microsecond=0))

    """Increments/decrements the month value of a date, until a month that matches the schedule is found.
    
    Args:
        cron_month_part (Part): The month 'Part' object.
        operation (Literal['add', 'subtract']): The function to call on date: 'add' or 'subtract'.
    """
    def _shift_month(self, cron_month_part: 'Part', operation: Literal['add', 'subtract']) -> None:
        while self.date.month not in cron_month_part.to_list():
            self.date = self._calc_months(self.date, 1, operation)

    """Increments/decrements the day value of a date, until a day that matches the schedule is found.

    Args:
        cron_day_part (Part): The days 'Part' object.
        operation (Literal['add', 'subtract']): The function to call on date: 'add' or 'subtract'.
    Returns:
        (boolean): Whether the month of the date was changed.
    """
    def _shift_day(self, cron_day_part: 'Part', cron_weekday_part: 'Part', operation: Literal['add', 'subtract']) -> bool:
        current_month = self.date.month
        while self.date.day not in cron_day_part.to_list() or \
                weekdays.get(self.date.strftime("%a")) not in cron_weekday_part.to_list():
            if operation == 'add':
                self.date = self.date + timedelta(days=+1)
                self.date = self.date.replace(hour=0, minute=0, second=0)
            else:
                self.date = self.date + timedelta(days=-1)
                self.date = self.date.replace(hour=23, minute=59, second=59)
            if current_month != self.date.month:
                return True
        return False

    """Increments/decrements the hour value of a date, until an hour that matches the schedule is found.

    Args:
        cron_hour_part (Part): The hours 'Part' object
        operation (Literal['add', 'subtract']): The function to call on date: 'add' or 'subtract'.
    Returns:
        (boolean): Whether the day of the date was changed
    """
    def _shift_hour(self, cron_hour_part: 'Part', operation: Literal['add', 'subtract']) -> bool:
        current_day = self.date.day
        while self.date.hour not in cron_hour_part.to_list():
            if operation == 'add':
                self.date = self.date + timedelta(hours=+1)
                self.date = self.date.replace(minute=00, second=0)
            else:
                self.date = self.date + timedelta(hours=-1)
                self.date = self.date.replace(minute=59, second=59)
            if current_day != self.date.day:
                return True
        return False

    """Increments/decrements the minute value of a date, until a minute that matches the schedule is found.

    Args:
        cron_minute_part (Part): The minutes 'Part' object.
        operation (Literal['add', 'subtract']): The function to call on date: 'add' or 'subtract'.

    Returns:
        (boolean): Whether the hour of the date was changed.
    """
    def _shift_minute(self, cron_minute_part: 'Part', operation: Literal['add', 'subtract']) -> bool:
        current_hour = self.date.hour
        while self.date.minute not in cron_minute_part.to_list():
            if operation == 'add':
                self.date = self.date + timedelta(minutes=+1)
                self.date = self.date.replace(second=0)
            else:
                self.date = self.date + timedelta(minutes=-1)
                self.date = self.date.replace(second=59)
            if current_hour != self.date.hour:
                return True
        return False

    """Static method to increment/decrement the month value of a datetime Object.
    
    Args:
        date (datetime): Date Object it increments/decrements the month value to.
        months (int): Number of months to add at the provided input date Object
        operation (Literal['add', 'subtract']): The function to call on date: 'add' or 'subtract'.
    Returns:
        (datetime): The input datetime object incremented or decremented.
    """
    @staticmethod
    def _calc_months(date: 'datetime', months: int, operation: Literal['add', 'subtract']) -> datetime:
        if operation == 'add':
            month = date.month - 1 + months
        else:
            month = date.month - 1 - months
        year = date.year + month // 12
        month = month % 12 + 1
        if operation == 'add':
            return date.replace(year=year, month=month, day=1, hour=00, minute=00, second=00)
        else:
            # Get the last day of the month
            max_month_day = calendar.monthrange(year, month)[1]
            return date.replace(year=year, month=month, day=max_month_day, hour=23, minute=59, second=59)



from setuptools import setup, find_packages

setup(
    name='cron-converter',
    version='0.4.3',
    license='MIT License',
    description='Cron string parser and scheduler for Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Andrea Salvatori',
    author_email='andrea.salvatori92@gmail.com',
    url='https://github.com/Sonic0/cron-converter',
    packages=['cron_converter', 'cron_converter/sub_modules'],
    keywords='cron',
    install_requires=['python-dateutil'],  # No requires
    include_package_data=True,
    extras_require={
            'test': ['unittest'],
        },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    python_requires='>=3.8',
)



from .sub_modules.part import Part
from .sub_modules.units import units
from .sub_modules.seeker import Seeker

from datetime import datetime
from functools import total_ordering

from typing import Optional, List, Union


@total_ordering
class Cron:
    """Creates an instance of Cron.

    Cron objects each represent a cron schedule.

    Attributes:
        options (dict): The options to use
    """
    def __init__(self, cron_string: str = None, options=None):
        self.options = options if bool(options) else dict()
        self.parts = None
        if cron_string:
            self.from_string(cron_string)

    def __str__(self) -> str:
        """Print directly the Cron Object"""
        return self.to_string()

    def __lt__(self, other) -> bool:
        """ This Cron object is lower than the other Cron.
        The comparison is made by 'total_ordering' comparing the number of Cron schedule times.
        """
        reordered_parts = self.parts[:3] + [self.parts[4], self.parts[3]]
        reordered_parts_other = other.parts[:3] + [other.parts[4], other.parts[3]]
        for part, other_part in zip(reversed(reordered_parts), reversed(reordered_parts_other)):
            if part < other_part:
                return True
        return False

    def __eq__(self, other) -> bool:
        """ This Cron object is equal to the other Cron.
        The comparison is made by 'total_ordering' comparing the number of Cron schedule times.
        """
        return all(part == other_part for part, other_part in zip(self.parts, other.parts))

    def from_string(self, cron_string: str) -> None:
        """Parses a cron string (minutes - hours - days - months - weekday)

        :param cron_string: (str) The cron string to parse. It has to be made up 5 parts.
        :raises ValueError: Incorrect length of the cron string.
        """
        if type(cron_string) != str:
            raise TypeError('Invalid cron string')
        self.parts = cron_string.strip().split()
        if len(self.parts) != 5:
            raise ValueError("Invalid cron string format")
        cron_parts = []
        for item, unit in zip(self.parts, units):
            part = Part(unit, self.options)
            part.from_string(item)
            cron_parts.append(part)

        self.parts: List[Part] = cron_parts

    def to_string(self) -> str:
        """Return the cron schedule as a string.

        :return: cron string (str) -> The cron schedule as a string.
        """
        if not self.parts:
            raise LookupError('No schedule found')
        return ' '.join(str(part) for part in self.parts)

    def from_list(self, cron_list: List[List[Union[str, int]]]):
        """Parses a 2-dimentional array of integers as a cron schedule.

        :param cron_list: (list of list) The 2-dimensional list to parse.
        :raises ValueError: Incorrect length of the cron list.
        """
        cron_parts = []
        if len(cron_list) != 5:
            raise ValueError(f'Invalid cron list')

        for cron_part_list, unit in zip(cron_list, units):
            part = Part(unit, self.options)
            part.from_list(cron_part_list)
            cron_parts.append(part)

        self.parts = cron_parts

    def to_list(self) -> List[List[int]]:
        """Returns the cron schedule as a 2-dimentional list of integers

        :return: schedule_list -> The cron schedule as a list.
        :raises LookupError: Empty Cron object.
        """
        if not self.parts:
            raise LookupError('No schedule found')
        schedule_list = []
        for part in self.parts:
            schedule_list.append(part.to_list())
        return schedule_list

    def schedule(self, start_date: Optional[datetime] = None, timezone_str: Optional[str] = None) -> Seeker:
        """Returns the time the schedule would run next.

        :param start_date: Optional. A datetime object. If not provided, date will be now in UTC.
                                     This param exclude 'timezone_str'.
        :param timezone_str: Optional. A timezone str('Europe/Rome', 'America/New_York', ...). Date will be now, but localized.
                                       If not provided, date will be now in UTC. This param exclude 'start_date'.
        :return: A schedule iterator.
        """
        return Seeker(self, start_date, timezone_str)
